import numpy as np

from yt.fields.derived_field import ValidateParameter
from yt.units import dimensions
from yt.units.yt_array import ustack
from yt.utilities.math_utils import get_sph_phi_component, get_sph_theta_component

from .field_plugin_registry import register_field_plugin


@register_field_plugin
def setup_magnetic_field_fields(registry, ftype="gas", slice_info=None):
    ds = registry.ds

    unit_system = ds.unit_system
    pc = registry.ds.units.physical_constants

    axis_names = registry.ds.coordinates.axis_order

    if (ftype, f"magnetic_field_{axis_names[0]}") not in registry:
        return

    u = registry[ftype, f"magnetic_field_{axis_names[0]}"].units

    def mag_factors(dims):
        if dims == dimensions.magnetic_field_cgs:
            return 4.0 * np.pi
        elif dims == dimensions.magnetic_field_mks:
            return ds.units.physical_constants.mu_0

    def _magnetic_field_strength(field, data):
        xm = f"relative_magnetic_field_{axis_names[0]}"
        ym = f"relative_magnetic_field_{axis_names[1]}"
        zm = f"relative_magnetic_field_{axis_names[2]}"

        B2 = (data[ftype, xm]) ** 2 + (data[ftype, ym]) ** 2 + (data[ftype, zm]) ** 2

        return np.sqrt(B2)

    registry.add_field(
        (ftype, "magnetic_field_strength"),
        sampling_type="local",
        function=_magnetic_field_strength,
        validators=[ValidateParameter("bulk_magnetic_field")],
        units=u,
    )

    def _magnetic_energy_density(field, data):
        B = data[ftype, "magnetic_field_strength"]
        return 0.5 * B * B / mag_factors(B.units.dimensions)

    registry.add_field(
        (ftype, "magnetic_energy_density"),
        sampling_type="local",
        function=_magnetic_energy_density,
        units=unit_system["pressure"],
    )

    registry.alias(
        (ftype, "magnetic_energy"),
        (ftype, "magnetic_energy_density"),
        deprecate=("4.0.0", "4.1.0"),
    )

    def _plasma_beta(field, data):
        return data[ftype, "pressure"] / data[ftype, "magnetic_energy_density"]

    registry.add_field(
        (ftype, "plasma_beta"), sampling_type="local", function=_plasma_beta, units=""
    )

    def _magnetic_pressure(field, data):
        return data[ftype, "magnetic_energy_density"]

    registry.add_field(
        (ftype, "magnetic_pressure"),
        sampling_type="local",
        function=_magnetic_pressure,
        units=unit_system["pressure"],
    )

    if registry.ds.geometry == "cartesian":

        def _magnetic_field_poloidal(field, data):
            normal = data.get_field_parameter("normal")

            Bfields = ustack(
                [
                    data[ftype, "relative_magnetic_field_x"],
                    data[ftype, "relative_magnetic_field_y"],
                    data[ftype, "relative_magnetic_field_z"],
                ]
            )

            theta = data["index", "spherical_theta"]
            phi = data["index", "spherical_phi"]

            return get_sph_theta_component(Bfields, theta, phi, normal)

        def _magnetic_field_toroidal(field, data):
            normal = data.get_field_parameter("normal")

            Bfields = ustack(
                [
                    data[ftype, "relative_magnetic_field_x"],
                    data[ftype, "relative_magnetic_field_y"],
                    data[ftype, "relative_magnetic_field_z"],
                ]
            )

            phi = data["index", "spherical_phi"]
            return get_sph_phi_component(Bfields, phi, normal)

    elif registry.ds.geometry == "cylindrical":

        def _magnetic_field_poloidal(field, data):
            bm = data.get_field_parameter("bulk_magnetic_field")
            r = data["index", "r"]
            z = data["index", "z"]
            d = np.sqrt(r * r + z * z)
            rax = axis_names.find("r")
            zax = axis_names.find("z")
            return (data[ftype, "magnetic_field_r"] - bm[rax]) * (r / d) + (
                data[ftype, "magnetic_field_z"] - bm[zax]
            ) * (z / d)

        def _magnetic_field_toroidal(field, data):
            ax = axis_names.find("theta")
            bm = data.get_field_parameter("bulk_magnetic_field")
            return data[ftype, "magnetic_field_theta"] - bm[ax]

    elif registry.ds.geometry == "spherical":

        def _magnetic_field_poloidal(field, data):
            ax = axis_names.find("theta")
            bm = data.get_field_parameter("bulk_magnetic_field")
            return data[ftype, "magnetic_field_theta"] - bm[ax]

        def _magnetic_field_toroidal(field, data):
            ax = axis_names.find("phi")
            bm = data.get_field_parameter("bulk_magnetic_field")
            return data[ftype, "magnetic_field_phi"] - bm[ax]

    else:

        # Unidentified geometry--set to None

        _magnetic_field_toroidal = None
        _magnetic_field_poloidal = None

    registry.add_field(
        (ftype, "magnetic_field_poloidal"),
        sampling_type="local",
        function=_magnetic_field_poloidal,
        units=u,
        validators=[
            ValidateParameter("normal"),
            ValidateParameter("bulk_magnetic_field"),
        ],
    )

    registry.add_field(
        (ftype, "magnetic_field_toroidal"),
        sampling_type="local",
        function=_magnetic_field_toroidal,
        units=u,
        validators=[
            ValidateParameter("normal"),
            ValidateParameter("bulk_magnetic_field"),
        ],
    )

    def _alfven_speed(field, data):
        B = data[ftype, "magnetic_field_strength"]
        return B / np.sqrt(mag_factors(B.units.dimensions) * data[ftype, "density"])

    registry.add_field(
        (ftype, "alfven_speed"),
        sampling_type="local",
        function=_alfven_speed,
        units=unit_system["velocity"],
    )

    def _mach_alfven(field, data):
        return data[ftype, "velocity_magnitude"] / data[ftype, "alfven_speed"]

    registry.add_field(
        (ftype, "mach_alfven"),
        sampling_type="local",
        function=_mach_alfven,
        units="dimensionless",
    )

    b_units = registry.ds.quan(1.0, u).units
    if dimensions.current_mks in b_units.dimensions.free_symbols:
        rm_scale = pc.qp.to("C", "SI") ** 3 / (4.0 * np.pi * pc.eps_0)
    else:
        rm_scale = pc.qp ** 3 / pc.clight
    rm_scale *= registry.ds.quan(1.0, "rad") / (
        2.0 * np.pi * pc.me ** 2 * pc.clight ** 3
    )
    rm_units = registry.ds.quan(1.0, "rad/m**2").units / unit_system["length"]

    def _rotation_measure(field, data):
        return (
            rm_scale
            * data[ftype, "magnetic_field_los"]
            * data[ftype, "El_number_density"]
        )

    registry.add_field(
        (ftype, "rotation_measure"),
        sampling_type="local",
        function=_rotation_measure,
        units=rm_units,
        validators=[ValidateParameter("axis", {"axis": [0, 1, 2]})],
    )


def setup_magnetic_field_aliases(registry, ds_ftype, ds_fields, ftype="gas"):
    r"""
    This routine sets up special aliases between dataset-specific magnetic
    fields and the default magnetic fields in yt so that unit conversions
    between different unit systems can be handled properly. This is only called
    from the `setup_fluid_fields` method (for grid dataset) or the
    `setup_gas_particle_fields` method (for particle dataset) of a frontend's
    :class:`FieldInfoContainer` instance.

    Parameters
    ----------
    registry : :class:`FieldInfoContainer`
        The field registry that these definitions will be installed into.
    ds_ftype : string
        The field type for the fields we're going to alias, e.g. "flash",
        "enzo", "athena", "PartType0", etc.
    ds_fields : list of strings or string
        The fields that will be aliased. For grid dataset, this should be a
        list of strings corresponding to the components of magnetic field. For
        particle dataset, this should be a single string corresponding to the
        vector magnetic field.
    ftype : string, optional
        The resulting field type of the fields. Default "gas".

    Examples
    --------
    >>> from yt.fields.magnetic_field import setup_magnetic_field_aliases
    >>> class PlutoFieldInfo(ChomboFieldInfo):
    ...     def setup_fluid_fields(self):
    ...         setup_magnetic_field_aliases(
    ...             self, "chombo", ["bx%s" % ax for ax in [1, 2, 3]]
    ...         )
    >>> class GizmoFieldInfo(GadgetFieldInfo):
    ...     def setup_gas_particle_fields(self):
    ...         setup_magnetic_field_aliases(
    ...             self, "PartType0", "MagneticField", ftype="PartType0"
    ...         )
    """
    unit_system = registry.ds.unit_system
    if isinstance(ds_fields, list):
        # If ds_fields is a list, we assume a grid dataset
        sampling_type = "local"
        ds_fields = [(ds_ftype, fd) for fd in ds_fields]
        ds_field = ds_fields[0]
    else:
        # Otherwise, we assume a particle dataset
        sampling_type = "particle"
        ds_field = (ds_ftype, ds_fields)
    if ds_field not in registry:
        return

    # Figure out the unit conversion to use
    if unit_system.base_units[dimensions.current_mks] is not None:
        to_units = unit_system["magnetic_field_mks"]
    else:
        to_units = unit_system["magnetic_field_cgs"]
    units = unit_system[to_units.dimensions]

    # Add fields
    if sampling_type in ["cell", "local"]:
        # Grid dataset case
        def mag_field(fd):
            def _mag_field(field, data):
                return data[fd].to(field.units)

            return _mag_field

        for ax, fd in zip(registry.ds.coordinates.axis_order, ds_fields):
            registry.add_field(
                (ftype, f"magnetic_field_{ax}"),
                sampling_type=sampling_type,
                function=mag_field(fd),
                units=units,
            )
    else:
        # Particle dataset case
        def mag_field(ax):
            def _mag_field(field, data):
                return data[ds_field][:, "xyz".index(ax)]

            return _mag_field

        for ax in registry.ds.coordinates.axis_order:
            fname = f"particle_magnetic_field_{ax}"
            registry.add_field(
                (ds_ftype, fname),
                sampling_type=sampling_type,
                function=mag_field(ax),
                units=units,
            )
            sph_ptypes = getattr(registry.ds, "_sph_ptypes", tuple())
            if ds_ftype in sph_ptypes:
                registry.alias((ftype, f"magnetic_field_{ax}"), (ds_ftype, fname))



from yt.funcs import is_sequence
from yt.utilities.logger import ytLogger as mylog

from .field_info_container import FieldInfoContainer
from .field_plugin_registry import register_field_plugin


class LocalFieldInfoContainer(FieldInfoContainer):
    def add_field(self, name, function, sampling_type, **kwargs):

        sampling_type = self._sanitize_sampling_type(
            sampling_type, kwargs.get("particle_type")
        )

        if isinstance(name, str) or not is_sequence(name):
            if sampling_type == "particle":
                ftype = "all"
            else:
                ftype = "gas"
            name = (ftype, name)

        override = kwargs.get("force_override", False)
        # Handle the case where the field has already been added.
        if not override and name in self:
            mylog.warning(
                "Field %s already exists. To override use `force_override=True`.",
                name,
            )

        return super().add_field(name, function, sampling_type, **kwargs)


# Empty FieldInfoContainer
local_fields = LocalFieldInfoContainer(None, [], None)

# we define two handles, pointing to the same function but documented differently
# yt.add_field() is meant to be used directly, while yt.derived_field is documented
# as a decorator.
add_field = local_fields.add_field


def derived_field(name, sampling_type, **kwargs):
    return add_field(name=name, function=None, sampling_type=sampling_type, **kwargs)


@register_field_plugin
def setup_local_fields(registry, ftype="gas", slice_info=None):
    # This is easy.  We just update with the contents of the local_fields field
    # info container, and since they are not mutable in any real way, we are
    # fine.
    # Note that we actually don't care about the ftype here.
    for f in local_fields:
        registry._show_field_errors.append(f)
    registry.update(local_fields)



import os
import warnings

import toml
from more_itertools import always_iterable

from yt._maintenance.deprecation import issue_deprecation_warning
from yt.utilities.configuration_tree import ConfigLeaf, ConfigNode

ytcfg_defaults = {}

ytcfg_defaults["yt"] = dict(
    serialize=False,
    only_deserialize=False,
    time_functions=False,
    colored_logs=False,
    suppress_stream_logging=False,
    stdout_stream_logging=False,
    log_level=20,
    inline=False,
    num_threads=-1,
    store_parameter_files=False,
    parameter_file_store="parameter_files.csv",
    maximum_stored_datasets=500,
    skip_dataset_cache=True,
    load_field_plugins=False,
    plugin_filename="my_plugins.py",
    parallel_traceback=False,
    pasteboard_repo="",
    reconstruct_index=True,
    test_storage_dir="/does/not/exist",
    test_data_dir="/does/not/exist",
    enzo_db="",
    notebook_password="",
    answer_testing_tolerance=3,
    answer_testing_bitwise=False,
    gold_standard_filename="gold311",
    local_standard_filename="local001",
    answer_tests_url="http://answers.yt-project.org/{1}_{2}",
    sketchfab_api_key="None",
    imagebin_api_key="e1977d9195fe39e",
    imagebin_upload_url="https://api.imgur.com/3/image",
    imagebin_delete_url="https://api.imgur.com/3/image/{delete_hash}",
    curldrop_upload_url="http://use.yt/upload",
    thread_field_detection=False,
    ignore_invalid_unit_operation_errors=False,
    chunk_size=1000,
    xray_data_dir="/does/not/exist",
    supp_data_dir="/does/not/exist",
    default_colormap="arbre",
    ray_tracing_engine="embree",
    internals=dict(
        within_testing=False,
        within_pytest=False,
        parallel=False,
        strict_requires=False,
        global_parallel_rank=0,
        global_parallel_size=1,
        topcomm_parallel_rank=0,
        topcomm_parallel_size=1,
        command_line=False,
    ),
)


def config_dir():
    config_root = os.environ.get(
        "XDG_CONFIG_HOME", os.path.join(os.path.expanduser("~"), ".config")
    )
    conf_dir = os.path.join(config_root, "yt")

    if not os.path.exists(conf_dir):
        try:
            os.makedirs(conf_dir)
        except OSError:
            warnings.warn("unable to create yt config directory")
    return conf_dir


def old_config_file():
    return os.path.join(config_dir(), "ytrc")


def old_config_dir():
    return os.path.join(os.path.expanduser("~"), ".yt")


# For backward compatibility, do not use these vars internally in yt
CONFIG_DIR = config_dir()
_OLD_CONFIG_FILE = old_config_file()


class YTConfig:
    def __init__(self, defaults=None):
        if defaults is None:
            defaults = {}
        self.config_root = ConfigNode(None)

    def get(self, section, *keys, callback=None):
        node_or_leaf = self.config_root.get(section, *keys)
        if isinstance(node_or_leaf, ConfigLeaf):
            if callback is not None:
                return callback(node_or_leaf)
            return node_or_leaf.value
        return node_or_leaf

    def get_most_specific(self, section, *keys, **kwargs):
        use_fallback = "fallback" in kwargs
        fallback = kwargs.pop("fallback", None)
        try:
            return self.config_root.get_deepest_leaf(section, *keys)
        except KeyError as err:
            if use_fallback:
                return fallback
            else:
                raise err

    def update(self, new_values, metadata=None):
        if metadata is None:
            metadata = {}
        self.config_root.update(new_values, metadata)

    def has_section(self, section):
        try:
            self.config_root.get_child(section)
            return True
        except KeyError:
            return False

    def add_section(self, section):
        self.config_root.add_child(section)

    def remove_section(self, section):
        if self.has_section(section):
            self.config_root.remove_child(section)
            return True
        else:
            return False

    def set(self, *args, metadata=None):
        section, *keys, value = args
        if metadata is None:
            metadata = {"source": "runtime"}
        self.config_root.upsert_from_list(
            [section] + list(keys), value, extra_data=metadata
        )

    def remove(self, *args):
        self.config_root.pop_leaf(args)

    def read(self, file_names):
        file_names_read = []
        for fname in always_iterable(file_names):
            if not os.path.exists(fname):
                continue
            metadata = {"source": f"file: {fname}"}
            self.update(toml.load(fname), metadata=metadata)
            file_names_read.append(fname)

        return file_names_read

    def write(self, file_handler):
        value = self.config_root.as_dict()
        config_as_str = toml.dumps(value)

        try:
            # Assuming file_handler has a write attribute
            file_handler.write(config_as_str)
        except AttributeError:
            # Otherwise we expect a path to a file
            with open(file_handler, mode="w") as fh:
                fh.write(config_as_str)

    @staticmethod
    def get_global_config_file():
        return os.path.join(config_dir(), "yt.toml")

    @staticmethod
    def get_local_config_file():
        return os.path.join(os.path.abspath(os.curdir), "yt.toml")

    def __setitem__(self, args, value):
        section, *keys = always_iterable(args)
        self.set(section, *keys, value, metadata=None)

    def __getitem__(self, key):
        section, *keys = always_iterable(key)
        return self.get(section, *keys)

    def __contains__(self, item):
        return item in self.config_root

    # Add support for IPython rich display
    # see https://ipython.readthedocs.io/en/stable/config/integrating.html
    def _repr_json_(self):
        return self.config_root._repr_json_()


_global_config_file = YTConfig.get_global_config_file()
_local_config_file = YTConfig.get_local_config_file()

if os.path.exists(old_config_file()):
    if os.path.exists(_global_config_file):
        issue_deprecation_warning(
            f"The configuration file {old_config_file()} is deprecated in "
            f"favor of {_global_config_file}. Currently, both are present. "
            "Please manually remove the deprecated one to silence "
            "this warning.",
            since="4.0.0",
            removal="4.1.0",
        )
    else:
        issue_deprecation_warning(
            f"The configuration file {_OLD_CONFIG_FILE} is deprecated. "
            f"Please migrate your config to {_global_config_file} by running: "
            "'yt config migrate'",
            since="4.0.0",
            removal="4.1.0",
        )


if not os.path.exists(_global_config_file):
    cfg = {"yt": {}}
    try:
        with open(_global_config_file, mode="w") as fd:
            toml.dump(cfg, fd)
    except OSError:
        warnings.warn("unable to write new config file")


# Load the config
ytcfg = YTConfig()
ytcfg.update(ytcfg_defaults, metadata={"source": "defaults"})

# Try loading the local config first, otherwise fall back to global config
if os.path.exists(_local_config_file):
    ytcfg.read(_local_config_file)
elif os.path.exists(_global_config_file):
    ytcfg.read(_global_config_file)



import os
from typing import Optional

from yt.utilities.logger import ytLogger as mylog

from ._mpl_imports import (
    FigureCanvasAgg,
    FigureCanvasPdf,
    FigureCanvasPS,
    FigureCanvasSVG,
)

AGG_FORMATS = [".png", ".jpg", ".jpeg", ".raw", ".rgba", ".tif", ".tiff"]
SUPPORTED_FORMATS = AGG_FORMATS + [".eps", ".ps", ".pdf", ".svg"]


def normalize_extension_string(s: str) -> str:
    return f".{s.lstrip('.')}"


def validate_image_name(filename, suffix: Optional[str] = None) -> str:
    """
    Build a valid image filename with a specified extension (default to png).
    The suffix parameter is ignored if the input filename has a valid extension already.
    Otherwise, suffix is appended to the filename, replacing any existing extension.
    """
    name, psuffix = os.path.splitext(filename)
    if psuffix in SUPPORTED_FORMATS:
        if suffix is not None:
            suffix = normalize_extension_string(suffix)
        if suffix in SUPPORTED_FORMATS and suffix != psuffix:
            mylog.warning(
                "Received two valid image formats '%s' (from `filename`) "
                "and '%s' (from `suffix`). The former is ignored.",
                psuffix,
                suffix,
            )
            return f"{name}{suffix}"
        return str(filename)

    if suffix is None:
        suffix = ".png"

    suffix = normalize_extension_string(suffix)

    if suffix not in SUPPORTED_FORMATS:
        raise ValueError("Unsupported file format '{suffix}'.")

    return f"{filename}{suffix}"


def get_canvas(figure, filename):

    name, suffix = os.path.splitext(filename)

    if not suffix:
        raise ValueError(
            f"Can not determine canvas class from filename '{filename}' "
            f"without an extension."
        )

    if suffix in AGG_FORMATS:
        return FigureCanvasAgg(figure)

    if suffix == ".pdf":
        return FigureCanvasPdf(figure)

    if suffix == ".svg":
        return FigureCanvasSVG(figure)

    if suffix in (".eps", ".ps"):
        return FigureCanvasPS(figure)

    raise ValueError(
        f"No matching canvas for filename '{filename}' with extension '{suffix}'."
    )
